import sqlite3
import sys
from abc import ABCMeta, abstractmethod
from itertools import chain
from typing import Generic, Optional, Tuple, Union, cast

if sys.version_info >= (3, 9):
    from collections.abc import Callable, Iterable, Mapping
else:
    from typing import Callable, Iterable, Mapping

from .base import KT, VT, SqliteCollectionBase, T, tidy_connection
from .dict import Dict
from .list import List
from .set import Set


class FactoryBase(Generic[T], metaclass=ABCMeta):
    def __init__(
        self,
        connection: Optional[Union[str, sqlite3.Connection]] = None,
        serializer: Optional[Callable[[T], bytes]] = None,
        deserializer: Optional[Callable[[bytes], T]] = None,
    ):
        self._connection = tidy_connection(connection)
        self._serializer = SqliteCollectionBase._default_serializer if serializer is None else serializer
        self._deserializer = SqliteCollectionBase._default_deserializer if deserializer is None else deserializer

    @property
    def connection(self) -> sqlite3.Connection:
        return self._connection

    @property
    def serializer(self) -> Callable[[T], bytes]:
        return self._serializer

    @property
    def deserializer(self) -> Callable[[bytes], T]:
        return self._deserializer

    @classmethod
    @abstractmethod
    def _get_container_class(
        cls,
    ) -> Callable[..., SqliteCollectionBase[T]]:
        ...


class SequenceFactoryBase(FactoryBase[T]):
    def create(self, __data: Optional[Iterable[T]] = None) -> SqliteCollectionBase[T]:
        if __data is None:
            return self._get_container_class()(
                connection=self.connection, serializer=self.serializer, deserializer=self.deserializer
            )
        return self._get_container_class()(
            data=__data, connection=self.connection, serializer=self.serializer, deserializer=self.deserializer
        )

    def __call__(self, __data: Optional[Iterable[T]] = None) -> SqliteCollectionBase[T]:
        return self.create(__data)


class SetFactory(SequenceFactoryBase[T]):
    @classmethod
    def _get_container_class(cls) -> Callable[..., Set[T]]:
        return Set[T]


class ListFactory(SequenceFactoryBase[T]):
    @classmethod
    def _get_container_class(cls) -> Callable[..., List[T]]:
        return List[T]


class DictFactory(FactoryBase[KT], Generic[KT, VT]):
    def __init__(
        self,
        connection: Optional[Union[str, sqlite3.Connection]] = None,
        key_serializer: Optional[Callable[[KT], bytes]] = None,
        key_deserializer: Optional[Callable[[bytes], KT]] = None,
        value_serializer: Optional[Callable[[VT], bytes]] = None,
        value_deserializer: Optional[Callable[[bytes], VT]] = None,
    ):
        super(DictFactory, self).__init__(
            connection=connection, serializer=key_serializer, deserializer=key_deserializer
        )
        self._value_serializer = (
            cast(Callable[[VT], bytes], self.key_serializer) if value_serializer is None else value_serializer
        )
        self._value_deserializer = (
            cast(Callable[[bytes], VT], self.key_deserializer) if value_deserializer is None else value_deserializer
        )

    @classmethod
    def _get_container_class(cls) -> Callable[..., Dict[KT, VT]]:
        return Dict[KT, VT]

    @property
    def key_serializer(self) -> Callable[[KT], bytes]:
        return self.serializer

    @property
    def key_deserializer(self) -> Callable[[bytes], KT]:
        return self.deserializer

    @property
    def value_serializer(self) -> Callable[[VT], bytes]:
        return self._value_serializer

    @property
    def value_deserializer(self) -> Callable[[bytes], VT]:
        return self._value_deserializer

    def create(
        self, __data: Optional[Union[Iterable[Tuple[KT, VT]], Mapping[KT, VT]]] = None, **kwargs: VT
    ) -> Dict[KT, VT]:
        if __data is None:
            if len(kwargs) == 0:
                return self._get_container_class()(
                    connection=self.connection,
                    key_serializer=self.key_serializer,
                    key_deserializer=self.key_deserializer,
                    value_serializer=self.value_serializer,
                    value_deserializer=self.value_deserializer,
                )
            return self._get_container_class()(
                data=kwargs,
                connection=self.connection,
                key_serializer=self.key_serializer,
                key_deserializer=self.key_deserializer,
                value_serializer=self.value_serializer,
                value_deserializer=self.value_deserializer,
            )
        return self._get_container_class()(
            data=chain(__data.items() if isinstance(__data, Mapping) else __data, kwargs.items()),
            connection=self.connection,
            key_serializer=self.key_serializer,
            key_deserializer=self.key_deserializer,
            value_serializer=self.value_serializer,
            value_deserializer=self.value_deserializer,
        )

    def __call__(
        self, __data: Optional[Union[Iterable[Tuple[KT, VT]], Mapping[KT, VT]]] = None, **kwargs: VT
    ) -> Dict[KT, VT]:
        return self.create(__data, **kwargs)



import itertools
import sqlite3
import sys
import warnings
from itertools import chain
from typing import (
    AbstractSet,
    Any,
    Generic,
    Optional,
    Tuple,
    TypeVar,
    Union,
    cast,
    overload,
)

if sys.version_info >= (3, 9):
    from collections.abc import (
        Callable,
        Iterable,
        Iterator,
        Mapping,
        MutableMapping,
        Reversible,
        Set,
        Sized,
    )
else:
    from typing import Callable, Iterable, Iterator, Mapping, MutableMapping, Set, Sized

    if sys.version_info >= (3, 8):
        from typing import Reversible

if sys.version_info >= (3, 9):
    from collections.abc import ItemsView as ItemsViewType
    from collections.abc import KeysView as KeysViewType
    from collections.abc import ValuesView as ValuesViewType
else:
    from typing import KeysView as KeysViewType
    from typing import ValuesView as ValuesViewType
    from typing import ItemsView as ItemsViewType

from . import RebuildStrategy
from .base import (
    _T,
    KT,
    VT,
    SqliteCollectionBase,
    T,
    _SqliteCollectionBaseDatabaseDriver,
    is_hashable,
)
from .set import Set as sc_Set

_KT_co = TypeVar("_KT_co", covariant=True)
_VT_co = TypeVar("_VT_co", covariant=True)


class _DictDatabaseDriver(_SqliteCollectionBaseDatabaseDriver):
    @classmethod
    def do_create_table(
        cls, table_name: str, container_type_nam: str, schema_version: str, cur: sqlite3.Cursor
    ) -> None:
        cur.execute(
            f"CREATE TABLE {table_name} ("
            "serialized_key BLOB NOT NULL UNIQUE, "
            "serialized_value BLOB NOT NULL, "
            "item_order INTEGER PRIMARY KEY)"
        )

    @classmethod
    def delete_single_record_by_serialized_key(
        cls, table_name: str, cur: sqlite3.Cursor, serialized_key: bytes
    ) -> None:
        cur.execute(f"DELETE FROM {table_name} WHERE serialized_key=?", (serialized_key,))

    @classmethod
    def delete_all_records(cls, table_name: str, cur: sqlite3.Cursor) -> None:
        cur.execute(f"DELETE FROM {table_name}")

    @classmethod
    def is_serialized_key_in(cls, table_name: str, cur: sqlite3.Cursor, serialized_key: bytes) -> bool:
        cur.execute(f"SELECT 1 FROM {table_name} WHERE serialized_key=?", (serialized_key,))
        return len(list(cur)) > 0

    @classmethod
    def get_serialized_value_by_serialized_key(
        cls, table_name: str, cur: sqlite3.Cursor, serialized_key: bytes
    ) -> Union[None, bytes]:
        cur.execute(
            f"SELECT serialized_value FROM {table_name} WHERE serialized_key=?",
            (serialized_key,),
        )
        res = cur.fetchone()
        if res is None:
            return None
        return cast(bytes, res[0])

    @classmethod
    def get_next_order(cls, table_name: str, cur: sqlite3.Cursor) -> int:
        cur.execute(f"SELECT MAX(item_order) FROM {table_name}")
        res = cur.fetchone()[0]
        if res is None:
            return 0
        return cast(int, res) + 1

    @classmethod
    def get_count(cls, table_name: str, cur: sqlite3.Cursor) -> int:
        cur.execute(f"SELECT COUNT(*) FROM {table_name}")
        res = cur.fetchone()
        return cast(int, res[0])

    @classmethod
    def get_serialized_keys(cls, table_name: str, cur: sqlite3.Cursor) -> Iterable[bytes]:
        cur.execute(f"SELECT serialized_key FROM {table_name} ORDER BY item_order")
        for res in cur:
            yield cast(bytes, res[0])

    @classmethod
    def insert_serialized_value_by_serialized_key(
        cls, table_name: str, cur: sqlite3.Cursor, serialized_key: bytes, serialized_value: bytes
    ) -> None:
        item_order = cls.get_next_order(table_name, cur)
        cur.execute(
            f"INSERT INTO {table_name} (serialized_key, serialized_value, item_order) VALUES (?, ?, ?)",
            (serialized_key, serialized_value, item_order),
        )

    @classmethod
    def update_serialized_value_by_serialized_key(
        cls, table_name: str, cur: sqlite3.Cursor, serialized_key: bytes, serialized_value: bytes
    ) -> None:
        cur.execute(
            f"UPDATE {table_name} SET serialized_value=? WHERE serialized_key=?",
            (serialized_value, serialized_key),
        )

    @classmethod
    def upsert(
        cls,
        table_name: str,
        cur: sqlite3.Cursor,
        serialized_key: bytes,
        serialized_value: bytes,
    ) -> None:
        if cls.is_serialized_key_in(table_name, cur, serialized_key):
            cls.update_serialized_value_by_serialized_key(table_name, cur, serialized_key, serialized_value)
        else:
            cls.insert_serialized_value_by_serialized_key(table_name, cur, serialized_key, serialized_value)

    @classmethod
    def get_last_serialized_item(cls, table_name: str, cur: sqlite3.Cursor) -> Tuple[bytes, bytes]:
        cur.execute(f"SELECT serialized_key, serialized_value FROM {table_name} ORDER BY item_order DESC LIMIT 1")
        return cast(Tuple[bytes, bytes], cur.fetchone())

    @classmethod
    def get_reversed_serialized_keys(cls, table_name: str, cur: sqlite3.Cursor) -> Iterable[bytes]:
        cur.execute(f"SELECT serialized_key FROM {table_name} ORDER BY item_order DESC")
        for res in cur:
            yield cast(bytes, res[0])

    @classmethod
    def get_serialized_values(cls, table_name: str, cur: sqlite3.Cursor) -> Iterable[bytes]:
        cur.execute(f"SELECT serialized_value FROM {table_name} ORDER BY item_order")
        for res in cur:
            yield cast(bytes, res[0])

    @classmethod
    def get_reversed_serialized_values(cls, table_name: str, cur: sqlite3.Cursor) -> Iterable[bytes]:
        cur.execute(f"SELECT serialized_value FROM {table_name} ORDER BY item_order DESC")
        for res in cur:
            yield cast(bytes, res[0])

    @classmethod
    def get_serialized_items(cls, table_name: str, cur: sqlite3.Cursor) -> Iterable[Tuple[bytes, bytes]]:
        cur.execute(f"SELECT serialized_key, serialized_value FROM {table_name} ORDER BY item_order")
        for res in cur:
            yield (cast(bytes, res[0]), cast(bytes, res[1]))


class _Dict(SqliteCollectionBase[KT], MutableMapping[KT, VT], Generic[KT, VT]):
    _driver_class = _DictDatabaseDriver

    def __init__(
        self,
        connection: Optional[Union[str, sqlite3.Connection]] = None,
        table_name: Optional[str] = None,
        key_serializer: Optional[Callable[[KT], bytes]] = None,
        key_deserializer: Optional[Callable[[bytes], KT]] = None,
        value_serializer: Optional[Callable[[VT], bytes]] = None,
        value_deserializer: Optional[Callable[[bytes], VT]] = None,
        serializer: Optional[Callable[[VT], bytes]] = None,
        deserializer: Optional[Callable[[bytes], VT]] = None,
        persist: bool = True,
        rebuild_strategy: RebuildStrategy = RebuildStrategy.CHECK_WITH_FIRST_ELEMENT,
        data: Optional[Union[Iterable[Tuple[KT, VT]], Mapping[KT, VT]]] = None,
    ) -> None:
        if serializer is not None:
            warnings.warn(
                "serializer argument is deprecated. use key_serializer or value_serializer instead",
                DeprecationWarning,
                stacklevel=2,
            )
        self._value_serializer = (
            value_serializer
            if value_serializer is not None
            else serializer
            if serializer is not None
            else cast(Callable[[VT], bytes], self._default_serializer if key_serializer is None else key_serializer)
        )
        if deserializer is not None:
            warnings.warn(
                "deserializer argument is deprecated. use key_deserializer or value_deserializer instead",
                DeprecationWarning,
                stacklevel=2,
            )
        self._value_deserializer = (
            value_deserializer
            if value_deserializer is not None
            else deserializer
            if deserializer is not None
            else cast(
                Callable[[bytes], VT], self._default_deserializer if key_deserializer is None else key_deserializer
            )
        )
        super(_Dict, self).__init__(
            connection=connection,
            table_name=table_name,
            serializer=key_serializer,
            deserializer=key_deserializer,
            persist=persist,
            rebuild_strategy=rebuild_strategy,
        )
        if data is not None:
            self.clear()
            self.update(data)

    @property
    def key_serializer(self) -> Callable[[KT], bytes]:
        return self._serializer

    @property
    def key_deserializer(self) -> Callable[[bytes], KT]:
        return self._deserializer

    @property
    def value_serializer(self) -> Callable[[VT], bytes]:
        return self._value_serializer

    @property
    def value_deserializer(self) -> Callable[[bytes], VT]:
        return self._value_deserializer

    @property
    def schema_version(self) -> str:
        return "0"

    def _rebuild_check_with_first_element(self) -> bool:
        cur = self.connection.cursor()
        cur.execute(f"SELECT serialized_key FROM {self.table_name} ORDER BY item_order LIMIT 1")
        res = cur.fetchone()
        if res is None:
            return False
        serialized_key = cast(bytes, res[0])
        key = self.deserialize_key(serialized_key)
        return serialized_key != self.serialize_key(key)

    def _do_rebuild(self) -> None:
        cur = self.connection.cursor()
        last_order = -1
        while last_order is not None:
            cur.execute(
                f"SELECT item_order FROM {self.table_name} WHERE item_order > ? ORDER BY item_order LIMIT 1",
                (last_order,),
            )
            res = cur.fetchone()
            if res is None:
                break
            i = res[0]
            cur.execute(
                f"SELECT serialized_key, serialized_value FROM {self.table_name} WHERE item_order=?",
                (i,),
            )
            serialized_key, serialized_value = cur.fetchone()
            cur.execute(
                f"UPDATE {self.table_name} SET serialized_key=?, serialized_value=? WHERE item_order=?",
                (
                    self.serialize_key(self.deserialize_key(serialized_key)),
                    self.serialize_value(self.deserialize_value(serialized_value)),
                    i,
                ),
            )
            last_order = i

    def serialize_key(self, key: KT) -> bytes:
        if not is_hashable(key):
            raise TypeError(f"unhashable type: '{type(key).__name__}'")
        return self.key_serializer(key)

    def deserialize_key(self, serialized_key: bytes) -> KT:
        return self.key_deserializer(serialized_key)

    def serialize_value(self, value: VT) -> bytes:
        return self.value_serializer(value)

    def deserialize_value(self, value: bytes) -> VT:
        return self.value_deserializer(value)

    def __delitem__(self, key: KT) -> None:
        serialized_key = self.serialize_key(key)
        cur = self.connection.cursor()
        if not self._driver_class.is_serialized_key_in(self.table_name, cur, serialized_key):
            raise KeyError(key)
        self._driver_class.delete_single_record_by_serialized_key(self.table_name, cur, serialized_key)
        self.connection.commit()

    def __getitem__(self, key: KT) -> VT:
        serialized_key = self.serialize_key(key)
        cur = self.connection.cursor()
        serialized_value = self._driver_class.get_serialized_value_by_serialized_key(
            self.table_name, cur, serialized_key
        )
        if serialized_value is None:
            raise KeyError(key)
        return self.deserialize_value(serialized_value)

    def __iter__(self) -> Iterator[KT]:
        cur = self.connection.cursor()
        for serialized_key in self._driver_class.get_serialized_keys(self.table_name, cur):
            yield self.deserialize_key(serialized_key)

    def __len__(self) -> int:
        cur = self.connection.cursor()
        return self._driver_class.get_count(self.table_name, cur)

    def __setitem__(self, key: KT, value: VT) -> None:
        serialized_key = self.serialize_key(key)
        cur = self.connection.cursor()
        serialized_value = self.serialize_value(value)
        self._driver_class.upsert(self.table_name, cur, serialized_key, serialized_value)
        self.connection.commit()

    def _create_volatile_copy(
        self,
        data: Optional[Mapping[KT, VT]] = None,
    ) -> "Dict[KT, VT]":

        return Dict[KT, VT](
            connection=self.connection,
            key_serializer=self.key_serializer,
            key_deserializer=self.key_deserializer,
            value_serializer=self.value_serializer,
            value_deserializer=self.value_deserializer,
            rebuild_strategy=RebuildStrategy.SKIP,
            persist=False,
            data=(self if data is None else data),
        )

    def copy(self) -> "Dict[KT, VT]":
        return self._create_volatile_copy()

    @classmethod
    def fromkeys(cls, iterable: Iterable[KT], value: Optional[VT]) -> "Dict[KT, VT]":
        raise NotImplementedError

    @overload
    def pop(self, k: KT) -> VT:
        ...

    @overload
    def pop(self, k: KT, default: Union[VT, T] = ...) -> Union[VT, T]:
        ...

    def pop(self, k: KT, default: Optional[Union[VT, object]] = None) -> Union[VT, object]:
        cur = self.connection.cursor()
        serialized_key = self.serialize_key(k)
        serialized_value = self._driver_class.get_serialized_value_by_serialized_key(
            self.table_name, cur, serialized_key
        )
        if serialized_value is None:
            if default is None:
                raise KeyError(k)
            return default
        self._driver_class.delete_single_record_by_serialized_key(self.table_name, cur, serialized_key)
        self.connection.commit()
        return self.deserialize_value(serialized_value)

    def popitem(self) -> Tuple[KT, VT]:
        cur = self.connection.cursor()
        serialized_item = self._driver_class.get_last_serialized_item(self.table_name, cur)
        if serialized_item is None:
            raise KeyError("popitem(): dictionary is empty")
        self._driver_class.delete_single_record_by_serialized_key(self.table_name, cur, serialized_item[0])
        self.connection.commit()
        return (
            self.deserialize_key(serialized_item[0]),
            self.deserialize_value(serialized_item[1]),
        )

    @overload
    def update(self, __other: Mapping[KT, VT], **kwargs: VT) -> None:
        ...

    @overload
    def update(self, __other: Iterable[Tuple[KT, VT]], **kwargs: VT) -> None:
        ...

    @overload
    def update(self, **kwargs: VT) -> None:
        ...

    def update(self, __other: Optional[Union[Iterable[Tuple[KT, VT]], Mapping[KT, VT]]] = None, **kwargs: VT) -> None:
        cur = self.connection.cursor()
        for k, v in chain(
            tuple() if __other is None else __other.items() if isinstance(__other, Mapping) else __other,
            cast(Mapping[KT, VT], kwargs).items(),
        ):
            self._driver_class.upsert(self.table_name, cur, self.serialize_key(k), self.serialize_value(v))
        self.connection.commit()

    def clear(self) -> None:
        cur = self.connection.cursor()
        self._driver_class.delete_all_records(self.table_name, cur)
        self.connection.commit()

    def __contains__(self, o: object) -> bool:
        return self._driver_class.is_serialized_key_in(
            self.table_name, self.connection.cursor(), self.serialize_key(cast(KT, o))
        )

    @overload
    def get(self, key: KT) -> Union[VT, None]:
        ...

    @overload
    def get(self, key: KT, default_value: Union[VT, T]) -> Union[VT, T]:
        ...

    def get(self, key: KT, default_value: Optional[Union[VT, object]] = None) -> Union[VT, None, object]:
        serialized_key = self.serialize_key(key)
        cur = self.connection.cursor()
        serialized_value = self._driver_class.get_serialized_value_by_serialized_key(
            self.table_name, cur, serialized_key
        )
        if serialized_value is None:
            return default_value
        return self.deserialize_value(serialized_value)

    def setdefault(self, key: KT, default: VT = None) -> VT:  # type: ignore
        serialized_key = self.serialize_key(key)
        cur = self.connection.cursor()
        serialized_value = self._driver_class.get_serialized_value_by_serialized_key(
            self.table_name, cur, serialized_key
        )
        if serialized_value is None:
            self._driver_class.insert_serialized_value_by_serialized_key(
                self.table_name, cur, serialized_key, self.serialize_value(default)
            )
            return default
        return self.deserialize_value(serialized_value)

    def keys(self) -> "KeysView[KT]":
        return KeysView[KT](cast(Dict[KT, VT], self))

    def values(self) -> "ValuesView[VT]":
        return ValuesView[VT](cast(Dict[KT, VT], self))

    def items(self) -> "ItemsView[KT, VT]":
        return ItemsView[KT, VT](cast(Dict[KT, VT], self))


if sys.version_info >= (3, 8):

    class _ReversibleDict(_Dict[KT, VT], Reversible[KT]):
        def __reversed__(self) -> Iterator[KT]:
            cur = self.connection.cursor()
            for serialized_key in self._driver_class.get_reversed_serialized_keys(self.table_name, cur):
                yield self.deserialize_key(serialized_key)


if sys.version_info >= (3, 9):

    class Dict(_ReversibleDict[KT, VT]):
        def __or__(self, other: Mapping[KT, VT]) -> "Dict[KT, VT]":
            tmp = Dict(
                connection=self.connection,
                key_serializer=self.key_serializer,
                key_deserializer=self.key_deserializer,
                value_serializer=self.value_serializer,
                value_deserializer=self.value_deserializer,
                persist=self.persist,
                data=self,
            )
            tmp |= other
            return tmp

        def __ior__(self, other: Mapping[KT, VT]) -> "Dict[KT, VT]":
            self.update(other)
            return self

elif sys.version_info >= (3, 8):

    class Dict(_ReversibleDict[KT, VT]):
        ...

else:

    class Dict(_Dict[KT, VT]):
        ...


class MappingView(Sized):
    def __init__(self, mapping: Dict[Any, Any]):
        self._parent = mapping

    def __len__(self) -> int:
        return len(self._parent)


class KeysView(MappingView, KeysViewType[_KT_co], Generic[_KT_co]):
    def __init__(self, mapping: Dict[_KT_co, Any]):
        super(KeysView, self).__init__(mapping)

    def __and__(self, o: Iterable[Any]) -> sc_Set[_KT_co]:  # type: ignore[override]
        return sc_Set[_KT_co](
            connection=self._parent.connection,
            serializer=self._parent.key_serializer,
            deserializer=self._parent.key_deserializer,
            persist=False,
            data=(d for d in o if d in self._parent),
        )

    def __rand__(self, o: Iterable[_T]) -> sc_Set[_T]:  # type: ignore[override]
        return sc_Set[_T](
            connection=self._parent.connection,
            serializer=self._parent.key_serializer,
            deserializer=self._parent.key_deserializer,
            persist=False,
            data=(d for d in o if d in self._parent),
        )

    def __contains__(self, o: object) -> bool:
        return o in self._parent

    def __iter__(self) -> Iterator[_KT_co]:
        return iter(self._parent)

    if sys.version_info >= (3, 8):

        def __reversed__(self) -> Iterator[_KT_co]:
            return reversed(self._parent)

    def __or__(self, o: Iterable[_T]) -> sc_Set[Union[_KT_co, _T]]:  # type: ignore[override]
        return sc_Set[Union[_KT_co, _T]](
            connection=self._parent.connection,
            serializer=self._parent.key_serializer,
            deserializer=self._parent.key_deserializer,
            persist=False,
            data=itertools.chain(self._parent, o),
        )

    def __ror__(self, o: Iterable[_T]) -> sc_Set[Union[_KT_co, _T]]:  # type: ignore[override]
        return self.__or__(o)

    def __sub__(self, o: Iterable[Any]) -> sc_Set[_KT_co]:  # type: ignore[override]
        return sc_Set[KT](
            connection=self._parent.connection,
            serializer=self._parent.key_serializer,
            deserializer=self._parent.key_deserializer,
            persist=False,
            data=self._parent,
        ).difference(o)

    def __rsub__(self, o: Iterable[_T]) -> sc_Set[_T]:  # type: ignore[override]
        return sc_Set[_T](
            connection=self._parent.connection,
            serializer=self._parent.key_serializer,
            deserializer=self._parent.key_deserializer,
            persist=False,
            data=o,
        ).difference(self._parent)

    def __xor__(self, o: Iterable[_T]) -> sc_Set[Union[_KT_co, _T]]:  # type: ignore[override]
        return sc_Set[Union[KT, _T]](
            connection=self._parent.connection,
            serializer=self._parent.key_serializer,
            deserializer=self._parent.key_deserializer,
            persist=False,
            data=self._parent,
        ).symmetric_difference(o)

    def __rxor__(self, o: Iterable[_T]) -> sc_Set[Union[_KT_co, _T]]:  # type: ignore[override]
        return self.__xor__(o)


class ValuesView(MappingView, ValuesViewType[_VT_co], Generic[_VT_co]):
    def __init__(self, mapping: Dict[Any, _VT_co]):
        super(ValuesView, self).__init__(mapping)

    def __contains__(self, o: object) -> bool:
        for v in self:
            if o == v:
                return True
        return False

    def __iter__(self) -> Iterator[_VT_co]:
        cur = self._parent.connection.cursor()
        for sv in self._parent._driver_class.get_serialized_values(self._parent.table_name, cur):
            yield self._parent.deserialize_value(sv)

    if sys.version_info >= (3, 8):

        def __reversed__(self) -> Iterator[_VT_co]:
            cur = self._parent.connection.cursor()
            for sv in self._parent._driver_class.get_reversed_serialized_values(self._parent.table_name, cur):
                yield self._parent.deserialize_value(sv)


class ItemsView(MappingView, ItemsViewType[_KT_co, _VT_co]):
    def _item_serializer(self, o: Tuple[_KT_co, _VT_co]) -> bytes:
        if not is_hashable(o[1]):
            raise TypeError(f"unhashable type: '{type(o[1]).__name__}'")
        return SqliteCollectionBase[Tuple[bytes, bytes]]._default_serializer(
            (self._parent.serialize_key(o[0]), self._parent.serialize_value(o[1]))
        )

    def _item_deserializer(self, o: bytes) -> Tuple[_KT_co, _VT_co]:
        sk, sv = SqliteCollectionBase[Tuple[bytes, bytes]]._default_deserializer(o)
        return (self._parent.deserialize_key(sk), self._parent.deserialize_value(sv))

    def __init__(self, mapping: Dict[_KT_co, _VT_co]) -> None:
        super(ItemsView, self).__init__(mapping)

    def __and__(self, o: Iterable[Any]) -> sc_Set[Tuple[_KT_co, _VT_co]]:  # type: ignore[override]
        tmp = sc_Set[Tuple[_KT_co, _VT_co]](
            connection=self._parent.connection,
            serializer=self._item_serializer,
            deserializer=self._item_deserializer,
            persist=False,
            data=iter(self),
        )
        return sc_Set[Tuple[_KT_co, _VT_co]](
            connection=self._parent.connection,
            serializer=self._item_serializer,
            deserializer=self._item_deserializer,
            persist=False,
            data=(cast(Tuple[_KT_co, _VT_co], d) for d in o if d in tmp),
        )

    def __rand__(self, o: Iterable[_T]) -> sc_Set[_T]:  # type: ignore[override]
        return cast(sc_Set[_T], self & o)

    def __contains__(self, o: object) -> bool:
        if not isinstance(o, tuple) or len(o) != 2 or o[0] not in self._parent:
            return False
        return bool(self._parent[o[0]] == o[1])

    def __iter__(self) -> Iterator[Tuple[_KT_co, _VT_co]]:
        cur = self._parent.connection.cursor()
        for sk, sv in self._parent._driver_class.get_serialized_items(self._parent.table_name, cur):
            yield self._parent.deserialize_key(sk), self._parent.deserialize_value(sv)

    if sys.version_info >= (3, 8):

        def __reversed__(self) -> Iterator[Tuple[_KT_co, _VT_co]]:
            ...

    def __or__(self, o: Iterable[_T]) -> sc_Set[Union[Tuple[_KT_co, _VT_co], _T]]:  # type: ignore[override]
        return sc_Set[Union[Tuple[_KT_co, _VT_co], _T]](
            connection=self._parent.connection,
            serializer=cast(Callable[[Union[Tuple[_KT_co, _VT_co], _T]], bytes], self._item_serializer),
            deserializer=cast(Callable[[bytes], Union[Tuple[_KT_co, _VT_co], _T]], self._item_deserializer),
            persist=False,
            data=itertools.chain(self, o),
        )

    def __ror__(self, o: Iterable[_T]) -> sc_Set[Union[Tuple[_KT_co, _VT_co], _T]]:  # type: ignore[override]
        return self | o

    def __sub__(self, o: Iterable[Any]) -> sc_Set[Tuple[_KT_co, _VT_co]]:  # type: ignore[override]
        tmp = sc_Set[Tuple[_KT_co, _VT_co]](
            connection=self._parent.connection,
            serializer=self._item_serializer,
            deserializer=self._item_deserializer,
            persist=False,
            data=iter(self),
        )
        tmp.difference_update(o)
        return tmp

    def __rsub__(self, o: Iterable[_T]) -> sc_Set[_T]:  # type: ignore[override]
        tmp = sc_Set[Tuple[_KT_co, _VT_co]](
            connection=self._parent.connection,
            serializer=self._item_serializer,
            deserializer=self._item_deserializer,
            persist=False,
            data=iter(self),
        )
        return sc_Set[_T](
            connection=self._parent.connection,
            persist=False,
            data=(d for d in o if d not in tmp),
        )

    def __xor__(self, o: Iterable[_T]) -> sc_Set[Union[Tuple[_KT_co, _VT_co], _T]]:  # type: ignore[override]
        tmp = sc_Set[Union[Tuple[_KT_co, _VT_co], _T]](
            connection=self._parent.connection,
            serializer=cast(Callable[[Union[Tuple[_KT_co, _VT_co], _T]], bytes], self._item_serializer),
            deserializer=cast(Callable[[bytes], Union[Tuple[_KT_co, _VT_co], _T]], self._item_deserializer),
            persist=False,
            data=iter(self),
        )
        tmp.symmetric_difference_update(o)
        return tmp

    def __rxor__(self, o: Iterable[_T]) -> sc_Set[Union[Tuple[_KT_co, _VT_co], _T]]:  # type: ignore[override]
        return self ^ o



__version__ = "0.9.4"
__author__ = "osoken"
__description__ = "Python collections that are backended by sqlite3 DB and are compatible with the built-in collections"
__email__ = "osoken.devel@outlook.jp"
__package_name__ = "sqlitecollections"


from .base import RebuildStrategy
from .dict import Dict
from .list import List
from .set import Set

__all__ = ["Dict", "List", "Set", "RebuildStrategy"]



import sqlite3
import sys
from itertools import count, repeat
from typing import Any, Optional, Tuple, Union, cast, overload

if sys.version_info >= (3, 9):
    from collections.abc import Callable, Iterable, Iterator, MutableSequence
else:
    from typing import Callable, Iterable, MutableSequence, Iterator

from . import RebuildStrategy
from .base import SqliteCollectionBase, T, _SqliteCollectionBaseDatabaseDriver


def _generate_indices_from_slice(l: int, s: slice) -> Iterator[int]:
    step = 1 if s.step is None else s.step
    if step == 0:
        raise ValueError("slice step cannot be zero")
    if step > 0:
        step = step
        start = min(max(0 if s.start is None else (s.start if s.start >= 0 else l + s.start), 0), l)
        stop = min(max(l if s.stop is None else (s.stop if s.stop >= 0 else l + s.stop), 0), l)
    else:
        step = step
        start = min(max(l - 1 if s.start is None else (s.start if s.start >= 0 else l + s.start), -1), l - 1)
        stop = min(max(-1 if s.stop is None else (s.stop if s.stop >= 0 else l + s.stop), -1), l)
    yield from range(start, stop, step)


class NoMoreElements(Exception):
    ...


class DifferentLengthDetected(Exception):
    def __init__(self, length1: int, length2: int) -> None:
        self.length1 = length1
        self.length2 = length2


def _consume_one_or_raise_no_more_elements(iter: Iterable[Any]) -> Any:
    for d in iter:
        return d
    raise NoMoreElements()


def _strict_zip(iter1: Iterable[Any], iter2: Iterable[Any]) -> Iterable[Tuple[Any, Any]]:
    element_count = 0
    while True:
        iter1_is_active = True
        try:
            next1 = _consume_one_or_raise_no_more_elements(iter1)
        except NoMoreElements:
            iter1_is_active = False

        iter2_is_active = True
        try:
            next2 = _consume_one_or_raise_no_more_elements(iter2)
        except NoMoreElements:
            iter2_is_active = False

        if iter1_is_active and iter2_is_active:
            yield (next1, next2)
            element_count += 1
        elif not iter1_is_active and not iter2_is_active:
            break
        else:
            iter1_unused_count = sum(d[0] for d in zip(repeat(1), iter1)) + int(iter1_is_active)
            iter2_unused_count = sum(d[0] for d in zip(repeat(1), iter2)) + int(iter2_is_active)
            raise DifferentLengthDetected(element_count + iter1_unused_count, element_count + iter2_unused_count)


class _ListDatabaseDriver(_SqliteCollectionBaseDatabaseDriver):
    @classmethod
    def do_create_table(
        cls, table_name: str, container_type_nam: str, schema_version: str, cur: sqlite3.Cursor
    ) -> None:
        cur.execute(f"CREATE TABLE {table_name} (serialized_value BLOB, item_index INTEGER PRIMARY KEY)")

    @classmethod
    def get_max_index_plus_one(cls, table_name: str, cur: sqlite3.Cursor) -> int:
        cur.execute(f"SELECT MAX(item_index) FROM {table_name}")
        res = cur.fetchone()
        if res[0] is None:
            return 0
        return cast(int, res[0]) + 1

    @classmethod
    def get_index_by_serialized_value(cls, table_name: str, cur: sqlite3.Cursor, serialized_value: bytes) -> int:
        cur.execute(f"SELECT item_index FROM {table_name} WHERE serialized_value = ? LIMIT 1", (serialized_value,))
        res = cur.fetchone()
        if res is None:
            return -1
        return cast(int, res[0])

    @classmethod
    def get_serialized_value_by_index(cls, table_name: str, cur: sqlite3.Cursor, index: int) -> Union[None, bytes]:
        if index < 0:
            l = cls.get_max_index_plus_one(table_name, cur)
            cur.execute(f"SELECT serialized_value FROM {table_name} WHERE item_index = ?", (l + index,))
        else:
            cur.execute(f"SELECT serialized_value FROM {table_name} WHERE item_index = ?", (index,))
        res = cur.fetchone()
        if res is None:
            return None
        return cast(bytes, res[0])

    @classmethod
    def tidy_indices(cls, table_name: str, cur: sqlite3.Cursor, cur2: sqlite3.Cursor, start: int = 0) -> None:
        cur.execute(f"SELECT item_index FROM {table_name} WHERE item_index >= ? ORDER BY item_index", (start,))
        for idx, d in zip(count(start), cur):
            idx_ = cast(int, d[0])
            if idx != idx_:
                cur2.execute(f"UPDATE {table_name} SET item_index = ? WHERE item_index = ?", (idx, idx_))

    @classmethod
    def delete_record_by_index(
        cls, table_name: str, cur: sqlite3.Cursor, index: int, length: Optional[int] = None
    ) -> Union[None, int]:
        _index = index
        if _index < 0:
            _length = length
            if _length is None:
                _length = cls.get_max_index_plus_one(table_name, cur)
            _index = _length + _index
        cur.execute(f"SELECT 1 FROM {table_name} WHERE item_index = ?", (_index,))
        if cur.fetchone() is None:
            return None
        cur.execute(f"DELETE FROM {table_name} WHERE item_index = ?", (_index,))
        return _index

    @classmethod
    def set_serialized_value_by_index(
        cls, table_name: str, cur: sqlite3.Cursor, serialized_value: bytes, index: int
    ) -> bool:
        _index = index
        l = cls.get_max_index_plus_one(table_name, cur)
        if _index < 0:
            _index = l + index
        if _index < 0 or l <= _index:
            return False
        cur.execute(f"UPDATE {table_name} SET serialized_value = ? WHERE item_index = ?", (serialized_value, _index))
        return True

    @classmethod
    def delete_all(cls, table_name: str, cur: sqlite3.Cursor) -> None:
        cur.execute(f"DELETE FROM {table_name}")

    @classmethod
    def add_record_by_serialized_value_and_index(
        cls, table_name: str, cur: sqlite3.Cursor, serialized_value: bytes, index: int
    ) -> None:
        cur.execute(f"INSERT INTO {table_name} (serialized_value, item_index) VALUES (?, ?)", (serialized_value, index))

    @classmethod
    def remap_index(cls, table_name: str, cur: sqlite3.Cursor, indices_map: Iterable[int]) -> None:
        l = cls.get_max_index_plus_one(table_name, cur)
        cur.execute(f"UPDATE {table_name} SET item_index = item_index - ?", (l,))
        cur.execute(
            f"UPDATE {table_name} SET item_index = CASE item_index {' '.join(['WHEN ? THEN ?' for _ in range(l)])} END",
            sum(((j - l, i) for i, j in enumerate(indices_map)), tuple()),
        )

    @classmethod
    def iter_serialized_value(cls, table_name: str, cur: sqlite3.Cursor) -> Iterable[bytes]:
        cur.execute(f"SELECT serialized_value FROM {table_name} ORDER BY item_index")
        for d in cur:
            yield cast(bytes, d[0])

    @classmethod
    def get_index_by_serialized_value_in_range(
        cls, table_name: str, cur: sqlite3.Cursor, serialized_value: bytes, normalized_start: int, normalized_stop: int
    ) -> Union[None, int]:
        cur.execute(
            f"SELECT item_index FROM {table_name} WHERE serialized_value = ? AND item_index >= ? AND item_index < ?",
            (serialized_value, normalized_start, normalized_stop),
        )
        res = cur.fetchone()
        if res is None:
            return None
        return cast(int, res[0])

    @classmethod
    def count_serialized_value(cls, table_name: str, cur: sqlite3.Cursor, serialized_value: bytes) -> int:
        cur.execute(
            f"SELECT COUNT(*) FROM {table_name} WHERE serialized_value = ?",
            (serialized_value,),
        )
        res = cur.fetchone()
        return cast(int, res[0])

    @classmethod
    def increment_indices(cls, table_name: str, cur: sqlite3.Cursor, start: int) -> None:
        idx = cls.get_max_index_plus_one(table_name, cur) - 1
        while idx >= start:
            cur.execute(f"UPDATE {table_name} SET item_index = ? WHERE item_index = ?", (idx + 1, idx))
            idx -= 1

    @classmethod
    def reverse_indices(cls, table_name: str, cur: sqlite3.Cursor) -> None:
        l = cls.get_max_index_plus_one(table_name, cur)
        cur.execute(f"UPDATE {table_name} SET item_index = -1 - item_index")
        cur.execute(f"UPDATE {table_name} SET item_index = item_index + ?", (l,))


class List(SqliteCollectionBase[T], MutableSequence[T]):
    _driver_class = _ListDatabaseDriver

    def __init__(
        self,
        connection: Optional[Union[str, sqlite3.Connection]] = None,
        table_name: Optional[str] = None,
        serializer: Optional[Callable[[T], bytes]] = None,
        deserializer: Optional[Callable[[bytes], T]] = None,
        persist: bool = True,
        rebuild_strategy: RebuildStrategy = RebuildStrategy.CHECK_WITH_FIRST_ELEMENT,
        data: Optional[Iterable[T]] = None,
    ) -> None:
        super(List, self).__init__(
            connection=connection,
            table_name=table_name,
            serializer=serializer,
            deserializer=deserializer,
            persist=persist,
            rebuild_strategy=rebuild_strategy,
        )
        if data is not None:
            self.clear()
            self.extend(data)

    def _do_rebuild(self) -> None:
        cur = self.connection.cursor()
        last_index = -1
        while last_index is not None:
            cur.execute(
                f"SELECT serialized_value, item_index FROM {self.table_name} WHERE item_index > ? ORDER BY item_index LIMIT 1",
                (last_index,),
            )
            res = cur.fetchone()
            if res is None:
                break
            cur.execute(
                f"UPDATE {self.table_name} SET serialized_value=? WHERE item_index=?",
                (self.serialize(self.deserialize(res[0])), res[1]),
            )
            last_index = res[1]

    def _rebuild_check_with_first_element(self) -> bool:
        cur = self.connection.cursor()
        cur.execute(f"SELECT serialized_value FROM {self.table_name} ORDER BY item_index LIMIT 1")
        res = cur.fetchone()
        if res is None:
            return False
        serialized_value = cast(bytes, res[0])
        value = self.deserialize(serialized_value)
        return serialized_value != self.serialize(value)

    @property
    def schema_version(self) -> str:
        return "0"

    def __delitem__(self, i: Union[int, slice]) -> None:
        cur = self.connection.cursor()
        cur2 = self.connection.cursor()
        if isinstance(i, int):
            deleted_index = self._driver_class.delete_record_by_index(self.table_name, cur, i)
            if deleted_index is None:
                raise IndexError("list assignment index out of range")
            self._driver_class.tidy_indices(self.table_name, cur, cur2, deleted_index)
            self.connection.commit()
            return
        reindexing_offset = None
        l = self._driver_class.get_max_index_plus_one(self.table_name, cur)
        for idx in _generate_indices_from_slice(l, i):
            self._driver_class.delete_record_by_index(self.table_name, cur, idx, l)
            if reindexing_offset is None or idx < reindexing_offset:
                reindexing_offset = idx
        if reindexing_offset is not None:
            self._driver_class.tidy_indices(self.table_name, cur, cur2, reindexing_offset)
        self.connection.commit()

    @overload
    def __getitem__(self, i: int) -> T:
        ...

    @overload
    def __getitem__(self, i: slice) -> "List[T]":
        ...

    def __getitem__(self, i: Union[int, slice]) -> "Union[T, List[T]]":
        cur = self.connection.cursor()
        if isinstance(i, int):
            serialized_value = self._driver_class.get_serialized_value_by_index(self.table_name, cur, i)
            if serialized_value is None:
                raise IndexError("list index out of range")
            return self.deserialize(serialized_value)
        l = self._driver_class.get_max_index_plus_one(self.table_name, cur)
        buf = self._create_volatile_copy([])
        bufcur = buf.connection.cursor()
        for next_idx, idx in enumerate(_generate_indices_from_slice(l, i)):
            buf._driver_class.add_record_by_serialized_value_and_index(
                buf.table_name,
                bufcur,
                cast(bytes, self._driver_class.get_serialized_value_by_index(self.table_name, cur, idx)),
                next_idx,
            )
        buf.connection.commit()
        return buf

    def _create_volatile_copy(self, data: Optional[Iterable[T]] = None) -> "List[T]":
        return List[T](
            connection=self.connection,
            serializer=self.serializer,
            deserializer=self.deserializer,
            rebuild_strategy=RebuildStrategy.SKIP,
            persist=False,
            data=(self if data is None else data),
        )

    def copy(self) -> "List[T]":
        return self._create_volatile_copy()

    def __setitem__(self, i: Union[int, slice], v: Union[T, Iterable[T]]) -> None:
        cur = self.connection.cursor()
        if isinstance(i, int):
            if not self._driver_class.set_serialized_value_by_index(
                self.table_name, cur, self.serialize(cast(T, v)), i
            ):
                raise IndexError("list assignment index out of range")
            self.connection.commit()
            return
        if not isinstance(v, Iterable):
            raise TypeError("must assign iterable to extended slice")
        l = self._driver_class.get_max_index_plus_one(self.table_name, cur)
        if i.step is None or i.step == 1:
            offset = min(max(0 if i.start is None else (i.start if i.start >= 0 else l + i.start), 0), l)
            del self[i]
            for idx, d in enumerate(v):
                self.insert(offset + idx, d)
            self.connection.commit()
        else:
            try:
                for idx, d in _strict_zip(_generate_indices_from_slice(l, i), v):
                    self._driver_class.set_serialized_value_by_index(self.table_name, cur, self.serialize(d), idx)
            except DifferentLengthDetected as e:
                raise ValueError(
                    f"attempt to assign sequence of size {e.length2} to extended slice of size {e.length1}"
                )
            self.connection.commit()
        return

    def __len__(self) -> int:
        cur = self.connection.cursor()
        return self._driver_class.get_max_index_plus_one(self.table_name, cur)

    def insert(self, i: int, v: T) -> None:
        cur = self.connection.cursor()
        index_ = i
        length = self._driver_class.get_max_index_plus_one(self.table_name, cur)
        if index_ < 0:
            index_ = length + index_
        index_ = max(0, min(length, index_))
        self._driver_class.increment_indices(self.table_name, cur, index_)
        self._driver_class.add_record_by_serialized_value_and_index(self.table_name, cur, self.serialize(v), index_)
        self.connection.commit()

    def __contains__(self, x: object) -> bool:
        cur = self.connection.cursor()
        serialized_value = self.serialize(cast(T, x))
        index = self._driver_class.get_index_by_serialized_value(self.table_name, cur, serialized_value)
        return index != -1

    def append(self, value: T) -> None:
        cur = self.connection.cursor()
        length = self._driver_class.get_max_index_plus_one(self.table_name, cur)
        self._driver_class.add_record_by_serialized_value_and_index(self.table_name, cur, self.serialize(value), length)
        self.connection.commit()

    def clear(self) -> None:
        cur = self.connection.cursor()
        self._driver_class.delete_all(self.table_name, cur)
        self.connection.commit()

    def extend(self, values: Iterable[T]) -> None:
        cur = self.connection.cursor()
        idx = self._driver_class.get_max_index_plus_one(self.table_name, cur)
        for v in values:
            self._driver_class.add_record_by_serialized_value_and_index(self.table_name, cur, self.serialize(v), idx)
            idx += 1
        self.connection.commit()

    def __iadd__(self, x: Iterable[T]) -> "List[T]":
        self.extend(x)
        return self

    def __add__(self, x: Iterable[T]) -> "List[T]":
        res = self.copy()
        res += x
        return res

    def __imul__(self, i: int) -> "List[T]":
        if not isinstance(i, int):
            raise TypeError(f"can't multiply sequence by non-int of type '{type(i).__name__}'")
        if i <= 0:
            self.clear()
            return self
        if i == 1:
            return self
        cur = self.connection.cursor()
        original_length = self._driver_class.get_max_index_plus_one(self.table_name, cur)
        for m in range(1, i):
            for j in range(original_length):
                serialized_value = cast(
                    bytes, self._driver_class.get_serialized_value_by_index(self.table_name, cur, j)
                )
                self._driver_class.add_record_by_serialized_value_and_index(
                    self.table_name, cur, serialized_value, m * original_length + j
                )
        self.connection.commit()
        return self

    def __mul__(self, i: int) -> "List[T]":
        res = self.copy()
        res *= i
        return res

    def index(self, value: Any, start: int = 0, stop: int = 0) -> int:
        cur = self.connection.cursor()
        length = None
        start_ = start
        if start_ < 0:
            length = self._driver_class.get_max_index_plus_one(self.table_name, cur)
            start_ = length + start_
        stop_ = stop
        if stop_ <= 0:
            if length is None:
                length = self._driver_class.get_max_index_plus_one(self.table_name, cur)
            stop_ = length + stop_
        serialized_value = self.serialize(cast(T, value))
        res = self._driver_class.get_index_by_serialized_value_in_range(
            self.table_name, cur, serialized_value, start_, stop_
        )
        if res is None:
            raise ValueError(f"'{value}' is not in list")
        return res

    def count(self, value: Any) -> int:
        cur = self.connection.cursor()
        return self._driver_class.count_serialized_value(self.table_name, cur, self.serialize(cast(T, value)))

    def pop(self, index: int = -1) -> T:
        cur = self.connection.cursor()
        cur2 = self.connection.cursor()
        length = self._driver_class.get_max_index_plus_one(self.table_name, cur)
        if length == 0:
            raise IndexError("pop from empty list")
        index_ = index
        if index_ < 0:
            index_ = length + index_
        if index_ < 0 or length <= index_:
            raise IndexError("pop index out of range")
        serialized_value = cast(bytes, self._driver_class.get_serialized_value_by_index(self.table_name, cur, index_))
        self._driver_class.delete_record_by_index(self.table_name, cur, index_)
        self._driver_class.tidy_indices(self.table_name, cur, cur2, index_)
        self.connection.commit()
        return self.deserialize(serialized_value)

    def sort(self, reverse: bool = False, key: Optional[Callable[[T], Any]] = None) -> None:
        key_ = (lambda x: x) if key is None else key
        cur = self.connection.cursor()
        buf = [
            (key_(self.deserialize(v)), i)
            for i, v in enumerate(self._driver_class.iter_serialized_value(self.table_name, cur))
        ]
        buf.sort(key=lambda x: x[0], reverse=reverse)  # type: ignore
        self._driver_class.remap_index(self.table_name, cur, [i[1] for i in buf])
        self.connection.commit()

    def reverse(self) -> None:
        cur = self.connection.cursor()
        self._driver_class.reverse_indices(self.table_name, cur)
        self.connection.commit()

    def remove(self, value: T) -> None:
        cur = self.connection.cursor()
        cur2 = self.connection.cursor()
        index = self._driver_class.get_index_by_serialized_value(self.table_name, cur, self.serialize(value))
        if index == -1:
            raise ValueError(f"'{value}' is not in list")
        self._driver_class.delete_record_by_index(self.table_name, cur, index)
        self._driver_class.tidy_indices(self.table_name, cur, cur2, index)
        self.connection.commit()
        return None



import sqlite3
import sys
from abc import ABCMeta, abstractmethod
from collections.abc import Hashable
from enum import Enum
from pickle import dumps, loads
from tempfile import NamedTemporaryFile
from types import TracebackType
from typing import IO, Generic, Optional, Type, TypeVar, Union, cast
from uuid import uuid4

from .logger import logger

if sys.version_info >= (3, 9):
    from contextlib import AbstractContextManager

    ContextManager = AbstractContextManager
    from collections.abc import Callable
else:
    from typing import Callable, ContextManager

T = TypeVar("T")
KT = TypeVar("KT")
VT = TypeVar("VT")
_T = TypeVar("_T")
_S = TypeVar("_S")


class RebuildStrategy(Enum):
    CHECK_WITH_FIRST_ELEMENT = 1
    ALWAYS = 2
    SKIP = 3


def sanitize_table_name(table_name: str, prefix: str) -> str:
    ret = "".join(c for c in table_name if c.isalnum() or c == "_")
    if len(ret) == 0:
        ret = create_random_name(prefix)
    if ret[0].isnumeric():
        ret = f"{prefix}_{ret}"
    if ret != table_name:
        logger.warning(f"The table name is changed to {ret} due to illegal characters")
    return ret


def create_random_name(prefix: str) -> str:
    return f"{prefix}_{str(uuid4()).replace('-', '')}"


def is_hashable(x: object) -> bool:
    return isinstance(x, Hashable)


def create_temporary_db_file() -> IO[bytes]:
    return NamedTemporaryFile(prefix="sc_", suffix=".db")


def create_tempfile_connection() -> sqlite3.Connection:
    return sqlite3.connect(create_temporary_db_file().name)


def tidy_connection(connection: Optional[Union[str, sqlite3.Connection]] = None) -> sqlite3.Connection:
    if connection is None:
        return create_tempfile_connection()
    elif isinstance(connection, str):
        return sqlite3.connect(connection)
    elif isinstance(connection, sqlite3.Connection):
        return connection
    else:
        raise TypeError(
            f"connection argument must be None or a string or a sqlite3.Connection, not '{type(connection)}'"
        )


class TemporaryTableContext(ContextManager[str]):
    def __init__(self, cur: sqlite3.Cursor, reference_table_name: str):
        self._cursor = cur
        self._reference_table_name = reference_table_name
        self._table_name = create_random_name("tmp")

    def __enter__(self) -> str:
        self._cursor.execute(
            f"CREATE TABLE {self._table_name} AS SELECT * FROM {self._reference_table_name} WHERE 0 = 1"
        )
        return self._table_name

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        excinst: Optional[BaseException],
        exctb: Optional[TracebackType],
    ) -> None:
        self._cursor.execute(f"DROP TABLE {self._table_name}")
        return None


class _SqliteCollectionBaseDatabaseDriver(metaclass=ABCMeta):
    @classmethod
    def initialize_metadata_table(cls, cur: sqlite3.Cursor) -> None:
        if not cls.is_metadata_table_initialized(cur):
            cls.do_initialize_metadata_table(cur)

    @classmethod
    def is_metadata_table_initialized(cls, cur: sqlite3.Cursor) -> bool:
        try:
            cur.execute("SELECT 1 FROM metadata LIMIT 1")
            _ = list(cur)
            return True
        except sqlite3.OperationalError as _:
            pass
        return False

    @classmethod
    def do_initialize_metadata_table(cls, cur: sqlite3.Cursor) -> None:
        cur.execute(
            """
            CREATE TABLE metadata (
                table_name TEXT PRIMARY KEY,
                schema_version TEXT NOT NULL,
                container_type TEXT NOT NULL,
                UNIQUE (table_name, container_type)
            )
            """
        )

    @classmethod
    def initialize_table(
        cls, table_name: str, container_type_name: str, schema_version: str, cur: sqlite3.Cursor
    ) -> None:
        if not cls.is_table_initialized(table_name, container_type_name, schema_version, cur):
            cls.do_create_table(table_name, container_type_name, schema_version, cur)
            cls.do_tidy_table_metadata(table_name, container_type_name, schema_version, cur)

    @classmethod
    def is_table_initialized(
        self, table_name: str, container_type_name: str, schema_version: str, cur: sqlite3.Cursor
    ) -> bool:
        try:
            cur.execute(
                "SELECT schema_version FROM metadata WHERE table_name=? AND container_type=?",
                (table_name, container_type_name),
            )
            buf = cur.fetchone()
            if buf is None:
                return False
            version = buf[0]
            if version != schema_version:
                return False
            cur.execute(f"SELECT 1 FROM {table_name} LIMIT 1")
            _ = list(cur)
            return True
        except sqlite3.OperationalError as _:
            pass
        return False

    @classmethod
    def do_tidy_table_metadata(
        cls, table_name: str, container_type_name: str, schema_version: str, cur: sqlite3.Cursor
    ) -> None:
        cur.execute(
            "INSERT INTO metadata (table_name, schema_version, container_type) VALUES (?, ?, ?)",
            (table_name, schema_version, container_type_name),
        )

    @classmethod
    @abstractmethod
    def do_create_table(
        cls, table_name: str, container_type_name: str, schema_version: str, cur: sqlite3.Cursor
    ) -> None:
        ...

    @classmethod
    def drop_table(cls, table_name: str, container_type_name: str, cur: sqlite3.Cursor) -> None:
        cur.execute(
            "DELETE FROM metadata WHERE table_name=? AND container_type=?",
            (table_name, container_type_name),
        )
        cur.execute(f"DROP TABLE {table_name}")

    @classmethod
    def alter_table_name(cls, table_name: str, new_table_name: str, cur: sqlite3.Cursor) -> None:
        cur.execute("UPDATE metadata SET table_name=? WHERE table_name=?", (new_table_name, table_name))
        cur.execute(f"ALTER TABLE {table_name} RENAME TO {new_table_name}")


class SqliteCollectionBase(Generic[T], metaclass=ABCMeta):
    _driver_class = _SqliteCollectionBaseDatabaseDriver

    @classmethod
    def _default_serializer(cls, x: T) -> bytes:
        return dumps(x, protocol=3)

    @classmethod
    def _default_deserializer(cls, x: bytes) -> T:
        return cast(T, loads(x))

    def __init__(
        self,
        connection: Optional[Union[str, sqlite3.Connection]] = None,
        table_name: Optional[str] = None,
        serializer: Optional[Callable[[T], bytes]] = None,
        deserializer: Optional[Callable[[bytes], T]] = None,
        persist: bool = True,
        rebuild_strategy: RebuildStrategy = RebuildStrategy.CHECK_WITH_FIRST_ELEMENT,
    ):
        super(SqliteCollectionBase, self).__init__()
        self._serializer = self._default_serializer if serializer is None else serializer
        self._deserializer = self._default_deserializer if deserializer is None else deserializer
        self._connection = tidy_connection(connection)
        self._persist = persist
        self._table_name = (
            sanitize_table_name(create_random_name(self.container_type_name), self.container_type_name)
            if table_name is None
            else sanitize_table_name(table_name, self.container_type_name)
        )
        self._initialize(rebuild_strategy=rebuild_strategy)

    def __del__(self) -> None:
        if not self.persist:
            cur = self.connection.cursor()
            self._driver_class.drop_table(self.table_name, self.container_type_name, cur)
            self.connection.commit()

    def _initialize(self, rebuild_strategy: RebuildStrategy) -> None:
        cur = self.connection.cursor()
        self._driver_class.initialize_metadata_table(cur)
        self._driver_class.initialize_table(self.table_name, self.container_type_name, self.schema_version, cur)
        if self._should_rebuild(rebuild_strategy):
            self._do_rebuild()
        self.connection.commit()

    def _should_rebuild(self, rebuild_strategy: RebuildStrategy) -> bool:
        if rebuild_strategy == RebuildStrategy.ALWAYS:
            return True
        if rebuild_strategy == RebuildStrategy.SKIP:
            return False
        return self._rebuild_check_with_first_element()

    @abstractmethod
    def _rebuild_check_with_first_element(self) -> bool:
        ...

    @abstractmethod
    def _do_rebuild(self) -> None:
        ...

    @property
    def persist(self) -> bool:
        return self._persist

    def set_persist(self, persist: bool) -> None:
        self._persist = persist

    @property
    def serializer(self) -> Callable[[T], bytes]:
        return self._serializer

    def serialize(self, x: T) -> bytes:
        return self.serializer(x)

    @property
    def deserializer(self) -> Callable[[bytes], T]:
        return self._deserializer

    def deserialize(self, blob: bytes) -> T:
        return self.deserializer(blob)

    @property
    def table_name(self) -> str:
        return self._table_name

    @table_name.setter
    def table_name(self, table_name: str) -> None:
        cur = self.connection.cursor()
        new_table_name = sanitize_table_name(table_name, self.container_type_name)
        if self._table_name != new_table_name:
            try:
                self._driver_class.alter_table_name(self.table_name, new_table_name, cur)
            except sqlite3.IntegrityError as e:
                raise ValueError(table_name)
            self._table_name = new_table_name

    @property
    def connection(self) -> sqlite3.Connection:
        return self._connection

    @property
    def container_type_name(self) -> str:
        return self.__class__.__name__

    @property
    @abstractmethod
    def schema_version(self) -> str:
        ...



import os

from setuptools import setup

from sqlitecollections import (
    __author__,
    __description__,
    __email__,
    __package_name__,
    __version__,
)

with open(os.path.join(os.path.dirname(os.path.abspath(__file__)), "README.md"), "r") as fin:
    __long_description__ = fin.read()

setup(
    name=__package_name__,
    version=__version__,
    author=__author__,
    author_email=__email__,
    license="MIT",
    url="https://github.com/osoken/sqlitecollections",
    description=__description__,
    long_description=__long_description__,
    long_description_content_type="text/markdown",
    packages=[__package_name__],
    project_urls={'Documentation': "https://osoken.github.io/sqlitecollections/"},
    install_requires=[],
    extras_require={
        "dev": ["flake8", "pytest", "black", "mypy", "tox", "isort"],
        "docs": ["mkdocs", "mkdocs-material", "mdx_include", "mkdocs-macros-plugin", "memory_profiler"],
    },
)



import sqlite3
import sys
from typing import AbstractSet, Any, Optional, Union, cast
from uuid import uuid4

if sys.version_info >= (3, 9):
    from collections.abc import Callable, Iterable, Iterator, MutableSet
else:
    from typing import Iterable, Iterator, MutableSet, Callable

from . import RebuildStrategy
from .base import (
    _S,
    _T,
    SqliteCollectionBase,
    T,
    TemporaryTableContext,
    _SqliteCollectionBaseDatabaseDriver,
    is_hashable,
)


class _SetDatabaseDriver(_SqliteCollectionBaseDatabaseDriver):
    @classmethod
    def do_create_table(
        cls, table_name: str, container_type_nam: str, schema_version: str, cur: sqlite3.Cursor
    ) -> None:
        cur.execute(f"CREATE TABLE {table_name} (serialized_value BLOB PRIMARY KEY)")

    @classmethod
    def delete_all(cls, table_name: str, cur: sqlite3.Cursor) -> None:
        cur.execute(f"DELETE FROM {table_name}")

    @classmethod
    def insert(cls, table_name: str, cur: sqlite3.Cursor, serialized_value: bytes) -> None:
        cur.execute(
            f"INSERT INTO {table_name} (serialized_value) VALUES (?)",
            (serialized_value,),
        )

    @classmethod
    def upsert(cls, table_name: str, cur: sqlite3.Cursor, serialized_value: bytes) -> None:
        if not cls.is_serialized_value_in(table_name, cur, serialized_value):
            cls.insert(table_name, cur, serialized_value)

    @classmethod
    def delete_by_serialized_value(cls, table_name: str, cur: sqlite3.Cursor, serialized_value: bytes) -> None:
        cur.execute(f"DELETE FROM {table_name} WHERE serialized_value = ?", (serialized_value,))

    @classmethod
    def is_serialized_value_in(cls, table_name: str, cur: sqlite3.Cursor, serialized_value: bytes) -> bool:
        cur.execute(f"SELECT 1 FROM {table_name} WHERE serialized_value=?", (serialized_value,))
        return len(list(cur)) > 0

    @classmethod
    def get_one_serialized_value(cls, table_name: str, cur: sqlite3.Cursor) -> Union[None, bytes]:
        cur.execute(f"SELECT serialized_value FROM {table_name} LIMIT 1")
        res = cur.fetchone()
        if res is None:
            return None
        return cast(bytes, res[0])

    @classmethod
    def get_count(cls, table_name: str, cur: sqlite3.Cursor) -> int:
        cur.execute(f"SELECT COUNT(*) FROM {table_name}")
        res = cur.fetchone()
        return cast(int, res[0])

    @classmethod
    def get_serialized_values(cls, table_name: str, cur: sqlite3.Cursor) -> Iterable[bytes]:
        cur.execute(f"SELECT serialized_value FROM {table_name}")
        for d in cur:
            yield cast(bytes, d[0])

    @classmethod
    def intersection_update_single(cls, table_name: str, cur: sqlite3.Cursor, data: Iterable[bytes]) -> None:
        with TemporaryTableContext(cur, table_name) as temp_table_name:
            for d in data:
                cls.upsert(temp_table_name, cur, d)
            cur.execute(
                f"DELETE FROM {table_name} WHERE NOT EXISTS (SELECT serialized_value FROM {temp_table_name} WHERE {table_name}.serialized_value = {temp_table_name}.serialized_value)"
            )

    @classmethod
    def difference_update_single(cls, table_name: str, cur: sqlite3.Cursor, data: Iterable[bytes]) -> None:
        for d in data:
            cls.delete_by_serialized_value(table_name, cur, d)

    @classmethod
    def union_update_single(cls, table_name: str, cur: sqlite3.Cursor, data: Iterable[bytes]) -> None:
        for d in data:
            cls.upsert(table_name, cur, d)

    @classmethod
    def symmetric_difference_update_single(
        cls, table_name: str, cur: sqlite3.Cursor, cur2: sqlite3.Cursor, data: Iterable[bytes]
    ) -> None:
        with TemporaryTableContext(cur, table_name) as temp_table_name:
            for d in data:
                cls.upsert(temp_table_name, cur, d)
            for serialized_value in cls.get_serialized_values(temp_table_name, cur2):
                if cls.is_serialized_value_in(table_name, cur, serialized_value):
                    cls.delete_by_serialized_value(table_name, cur, serialized_value)
                else:
                    cls.insert(table_name, cur, serialized_value)

    @classmethod
    def is_proper_superset(
        cls, table_name: str, cur: sqlite3.Cursor, cur2: sqlite3.Cursor, data: Iterable[bytes]
    ) -> bool:
        with TemporaryTableContext(cur, table_name) as temp_table_name:
            for d in data:
                if not cls.is_serialized_value_in(table_name, cur2, d):
                    return False
                cls.upsert(temp_table_name, cur2, d)
            return cls.get_count(temp_table_name, cur2) < cls.get_count(table_name, cur2)

    @classmethod
    def is_proper_subset(
        cls, table_name: str, cur: sqlite3.Cursor, cur2: sqlite3.Cursor, data: Iterable[bytes]
    ) -> bool:
        is_proper = False
        with TemporaryTableContext(cur, table_name) as temp_table_name:
            for d in data:
                if not cls.is_serialized_value_in(table_name, cur2, d):
                    is_proper = True
                else:
                    cls.upsert(temp_table_name, cur2, d)
            return is_proper and cls.get_count(temp_table_name, cur2) == cls.get_count(table_name, cur2)


class Set(SqliteCollectionBase[T], MutableSet[T]):
    _driver_class = _SetDatabaseDriver

    def __init__(
        self,
        connection: Optional[Union[str, sqlite3.Connection]] = None,
        table_name: Optional[str] = None,
        serializer: Optional[Callable[[T], bytes]] = None,
        deserializer: Optional[Callable[[bytes], T]] = None,
        persist: bool = True,
        rebuild_strategy: RebuildStrategy = RebuildStrategy.CHECK_WITH_FIRST_ELEMENT,
        data: Optional[Iterable[T]] = None,
    ) -> None:
        super(Set, self).__init__(
            connection=connection,
            table_name=table_name,
            serializer=serializer,
            deserializer=deserializer,
            persist=persist,
            rebuild_strategy=rebuild_strategy,
        )
        if data is not None:
            self.clear()
            self.update(data)

    def __contains__(self, value: object) -> bool:
        cur = self.connection.cursor()
        serialized_value = self.serialize(cast(T, value))
        return self._driver_class.is_serialized_value_in(self.table_name, cur, serialized_value)

    def __iter__(self) -> Iterator[T]:
        cur = self.connection.cursor()
        for d in self._driver_class.get_serialized_values(self.table_name, cur):
            yield self.deserialize(d)

    def __len__(self) -> int:
        cur = self.connection.cursor()
        return self._driver_class.get_count(self.table_name, cur)

    def _do_rebuild(self) -> None:
        cur = self.connection.cursor()
        backup_table_name = f'bk_{self.container_type_name}_{str(uuid4()).replace("-", "")}'
        cur.execute(f"CREATE TABLE {backup_table_name} AS SELECT * FROM {self.table_name}")
        cur.execute(f"DELETE FROM {self.table_name}")
        iter_old_records = self.connection.cursor()
        iter_old_records.execute(f"SELECT serialized_value FROM {backup_table_name}")
        delete_old_records = self.connection.cursor()
        insert_new_records = self.connection.cursor()
        for d in iter_old_records:
            insert_new_records.execute(
                f"INSERT INTO {self.table_name} (serialized_value) VALUES (?)",
                (self.serialize(self.deserialize(d[0])),),
            )
            delete_old_records.execute(f"DELETE FROM {backup_table_name} WHERE serialized_value = ?", d)
        cur.execute(f"DROP TABLE {backup_table_name}")

    def _rebuild_check_with_first_element(self) -> bool:
        cur = self.connection.cursor()
        cur.execute(f"SELECT serialized_value FROM {self.table_name} LIMIT 1")
        res = cur.fetchone()
        if res is None:
            return False
        serialized_value = cast(bytes, res[0])
        value = self.deserialize(serialized_value)
        return serialized_value != self.serialize(value)

    def serialize(self, value: T) -> bytes:
        if not is_hashable(value):
            raise TypeError(f"unhashable type: '{type(value).__name__}'")
        return self.serializer(value)

    def add(self, value: T) -> None:
        serialized_value = self.serialize(value)
        cur = self.connection.cursor()
        self._driver_class.upsert(self.table_name, cur, serialized_value)
        self.connection.commit()

    def clear(self) -> None:
        cur = self.connection.cursor()
        self._driver_class.delete_all(self.table_name, cur)
        self.connection.commit()

    def discard(self, value: T) -> None:
        cur = self.connection.cursor()
        self._driver_class.delete_by_serialized_value(self.table_name, cur, self.serialize(value))
        self.connection.commit()

    def remove(self, value: T) -> None:
        cur = self.connection.cursor()
        serialized_value = self.serialize(value)
        if not self._driver_class.is_serialized_value_in(self.table_name, cur, serialized_value):
            raise KeyError(value)
        self._driver_class.delete_by_serialized_value(self.table_name, cur, serialized_value)
        self.connection.commit()

    def pop(self) -> T:
        cur = self.connection.cursor()
        serialized_value = self._driver_class.get_one_serialized_value(self.table_name, cur)
        if serialized_value is None:
            raise KeyError("'pop from an empty set'")
        self._driver_class.delete_by_serialized_value(self.table_name, cur, serialized_value)
        self.connection.commit()
        return self.deserialize(serialized_value)

    @property
    def schema_version(self) -> str:
        return "0"

    def issubset(self, other: Iterable[T]) -> bool:
        return len(self) == len(self.intersection(other))

    def __lt__(self, other: AbstractSet[T]) -> bool:
        if len(self) >= len(other):
            return False
        for d in self:
            if d not in other:
                return False
        return True

    def __le__(self, other: AbstractSet[T]) -> bool:
        for d in self:
            if d not in other:
                return False
        return True

    def intersection(self, *others: Iterable[T]) -> "Set[T]":
        res = self.copy()
        res.intersection_update(*others)
        return res

    def intersection_update(self, *others: Iterable[T]) -> None:
        cur = self.connection.cursor()
        for other in others:
            self._driver_class.intersection_update_single(self.table_name, cur, (self.serialize(d) for d in other))
        self.connection.commit()

    def issuperset(self, other: Iterable[T]) -> bool:
        cur = self.connection.cursor()
        for d in other:
            if not self._driver_class.is_serialized_value_in(self.table_name, cur, self.serialize(d)):
                return False
        return True

    def __gt__(self, other: AbstractSet[T]) -> bool:
        return self._driver_class.is_proper_superset(
            self.table_name, self.connection.cursor(), self.connection.cursor(), (self.serialize(d) for d in other)
        )

    def __ge__(self, other: AbstractSet[T]) -> bool:
        return self.issuperset(other)

    def union(self, *others: Iterable[T]) -> "Set[T]":
        res = self.copy()
        res.update(*others)
        return res

    def update(self, *others: Iterable[T]) -> None:
        cur = self.connection.cursor()
        for other in others:
            self._driver_class.union_update_single(self.table_name, cur, (self.serialize(d) for d in other))
        self.connection.commit()

    def isdisjoint(self, other: Iterable[T]) -> bool:
        cur = self.connection.cursor()
        for d in other:
            if self._driver_class.is_serialized_value_in(self.table_name, cur, self.serialize(d)):
                return False
        return True

    @classmethod
    def _from_iterable(cls, *args: T) -> "Set[T]":
        raise NotImplementedError

    def __or__(self, s: AbstractSet[_T]) -> "Set[T]":
        return self.union(cast(Iterable[T], s))

    def __and__(self, s: AbstractSet[Any]) -> "Set[T]":
        return self.intersection(cast(Iterable[T], s))

    def difference(self, *others: Iterable[T]) -> "Set[T]":
        res = self.copy()
        res.difference_update(*others)
        return res

    def difference_update(self, *others: Iterable[T]) -> None:
        cur = self.connection.cursor()
        for other in others:
            self._driver_class.difference_update_single(self.table_name, cur, (self.serialize(d) for d in other))
        self.connection.commit()

    def _create_volatile_copy(self, data: Optional[Iterable[T]] = None) -> "Set[T]":
        return Set[T](
            connection=self.connection,
            serializer=self.serializer,
            deserializer=self.deserializer,
            rebuild_strategy=RebuildStrategy.SKIP,
            persist=False,
            data=data if data is not None else self,
        )

    def copy(self) -> "Set[T]":
        return self._create_volatile_copy()

    def __sub__(self, s: AbstractSet[Any]) -> "Set[T]":
        return self.difference(cast(Iterable[T], s))

    def symmetric_difference(self, *others: Iterable[T]) -> "Set[T]":
        res = self.copy()
        res.symmetric_difference_update(*others)
        return res

    def symmetric_difference_update(self, *others: Iterable[T]) -> None:
        cur = self.connection.cursor()
        cur2 = self.connection.cursor()
        for other in others:
            self._driver_class.symmetric_difference_update_single(
                self.table_name, cur, cur2, (self.serialize(d) for d in other)
            )
        self.connection.commit()

    def __xor__(self, s: AbstractSet[_T]) -> "Set[T]":
        return self.symmetric_difference(cast(Iterable[T], s))

    def __ixor__(self, s: AbstractSet[_S]) -> "Set[Union[_T, T]]":
        self.symmetric_difference_update(cast(Iterable[T], s))
        return cast(Set[Union[_T, T]], self)

    def __isub__(self, s: AbstractSet[Any]) -> "Set[T]":
        self.difference_update(cast(Iterable[T], s))
        return self

    def __iand__(self, s: AbstractSet[Any]) -> "Set[T]":
        self.intersection_update(cast(Iterable[T], s))
        return self

    def __ior__(self, s: AbstractSet[_S]) -> "Set[Union[_T, T]]":
        self.update(cast(Iterable[T], s))
        return cast(Set[Union[_T, T]], self)
